from enum import Enum
from typing import Union, Literal, Any

from pydantic import BaseModel, Field

from .invoice import StripeInvoiceEventData
from .price import StripePriceEventData
from .product import StripeProductEventData
from .subscription import StripeSubscriptionEventData


class EventType(str, Enum):
    """See: https://stripe.com/docs/api/events/types"""

    CUSTOMER_UPDATED = 'customer.updated'

    CUSTOMER_SUBSCRIPTION_CREATED = 'customer.subscription.created'
    CUSTOMER_SUBSCRIPTION_UPDATED = 'customer.subscription.updated'
    CUSTOMER_SUBSCRIPTION_DELETED = 'customer.subscription.deleted'

    INVOICE_CREATED = 'invoice.created'
    INVOICE_FINALIZED = 'invoice.finalized'
    INVOICE_PAYMENT_SUCCEEDED = 'invoice.payment_succeeded'
    INVOICE_PAYMENT_FAILED = 'invoice.payment_failed'
    INVOICE_PAID = 'invoice.paid'

    INVOICEITEM_CREATED = 'invoiceitem.created'

    PRODUCT_CREATED = 'product.created'
    PRODUCT_UPDATED = 'product.updated'
    PRODUCT_DELETED = 'product.deleted'

    PRICE_DELETED = 'price.deleted'
    PRICE_UPDATED = 'price.updated'
    PRICE_CREATED = 'price.created'


class StripeEventRequest(BaseModel):
    """Based on: https://stripe.com/docs/api/events/object#event_object-request"""
    id: str = None
    idempotency_key: str = None


class StripeBaseEvent(BaseModel):
    """
    Based on https://stripe.com/docs/api/events/object
    This is the base event template for more specific Stripe event classes
    """
    id: str
    api_version: str
    request: StripeEventRequest
    data: Any  # overwrite this attribute when inheriting
    type: Literal[str]  # overwrite this attribute when inheriting


class StripeInvoiceEvent(StripeBaseEvent):
    data: StripeInvoiceEventData
    type: Literal[
        EventType.INVOICE_PAID,
        EventType.INVOICE_CREATED,
        EventType.INVOICE_PAID,
        EventType.INVOICE_PAYMENT_FAILED
    ]


class StripeSubscriptionEvent(StripeBaseEvent):
    data: StripeSubscriptionEventData
    type: Literal[
        EventType.CUSTOMER_SUBSCRIPTION_DELETED,
        EventType.CUSTOMER_SUBSCRIPTION_UPDATED,
        EventType.CUSTOMER_SUBSCRIPTION_CREATED
    ]


class StripeProductEvent(StripeBaseEvent):
    data: StripeProductEventData
    type: Literal[
        EventType.PRODUCT_UPDATED,
        EventType.PRODUCT_CREATED,
        EventType.PRODUCT_DELETED
    ]


class StripePriceEvent(StripeBaseEvent):
    data: StripePriceEventData
    type: Literal[
        EventType.PRICE_CREATED,
        EventType.PRICE_UPDATED,
        EventType.PRICE_DELETED
    ]


class StripeEvent(BaseModel):
    # Add event classes to this attribute as they are implemented, more specific types first.
    # see https://pydantic-docs.helpmanual.io/usage/types/#discriminated-unions-aka-tagged-unions
    event: Union[
        StripeSubscriptionEvent,
        StripeInvoiceEvent,
        StripeProductEvent,
        StripePriceEvent,
        StripeBaseEvent,  # needed here so unimplemented event types can pass through validation
    ] = Field(discriminator='type')



from pydantic import ValidationError
from rest_framework.request import Request

from drf_stripe.settings import drf_stripe_settings
from drf_stripe.stripe_api.api import stripe_api as stripe
from drf_stripe.stripe_models.event import EventType
from drf_stripe.stripe_models.event import StripeEvent
from .customer_subscription import _handle_customer_subscription_event_data
from .price import _handle_price_event_data
from .product import _handle_product_event_data


def handle_stripe_webhook_request(request):
    event = _make_webhook_event_from_request(request)
    handle_webhook_event(event)


def _make_webhook_event_from_request(request: Request):
    """
    Given a Rest Framework request, construct a webhook event.

    :param event: event from Stripe Webhook, defaults to None. Used for test.
    """

    return stripe.Webhook.construct_event(
        payload=request.body,
        sig_header=request.META['HTTP_STRIPE_SIGNATURE'],
        secret=drf_stripe_settings.STRIPE_WEBHOOK_SECRET)


def _handle_event_type_validation_error(err: ValidationError):
    """
    Handle Pydantic ValidationError raised when parsing StripeEvent,
    ignores the error if it is caused by unimplemented event.type;
    Otherwise, raise the error.
    """
    event_type_error = False

    for error in err.errors():
        error_loc = error['loc']
        if error_loc[0] == 'event' and error_loc[1] == 'type':
            event_type_error = True
            break

    if event_type_error is False:
        raise err


def handle_webhook_event(event):
    """Perform actions given Stripe Webhook event data."""

    try:
        e = StripeEvent(event=event)
    except ValidationError as err:
        _handle_event_type_validation_error(err)
        return

    event_type = e.event.type

    if event_type is EventType.CUSTOMER_SUBSCRIPTION_CREATED:
        _handle_customer_subscription_event_data(e.event.data)

    elif event_type is EventType.CUSTOMER_SUBSCRIPTION_UPDATED:
        _handle_customer_subscription_event_data(e.event.data)

    elif event_type is EventType.CUSTOMER_SUBSCRIPTION_DELETED:
        _handle_customer_subscription_event_data(e.event.data)


    elif event_type is EventType.PRODUCT_CREATED:
        _handle_product_event_data(e.event.data)

    elif event_type is EventType.PRODUCT_UPDATED:
        _handle_product_event_data(e.event.data)

    elif event_type is EventType.PRODUCT_DELETED:
        _handle_product_event_data(e.event.data)


    elif event_type is EventType.PRICE_CREATED:
        _handle_price_event_data(e.event.data)

    elif event_type is EventType.PRICE_UPDATED:
        _handle_price_event_data(e.event.data)

    elif event_type is EventType.PRICE_DELETED:
        _handle_price_event_data(e.event.data)
   



from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from stripe.error import StripeError

from drf_stripe.models import SubscriptionItem, Product, Price, Subscription
from drf_stripe.stripe_api.checkout import stripe_api_create_checkout_session
from drf_stripe.stripe_api.customers import get_or_create_stripe_user


class SubscriptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Subscription
        fields = (
            "subscription_id", "period_start", "period_end", "status", "cancel_at", "cancel_at_period_end",
            "trial_start", "trial_end"
        )


class SubscriptionItemSerializer(serializers.ModelSerializer):
    """Serializes SubscriptionItem model with attributes pulled from related Subscription instance"""
    product_id = serializers.CharField(source="price.product.product_id")
    product_name = serializers.CharField(source="price.product.name")
    product_description = serializers.CharField(source="price.product.description")
    price_id = serializers.CharField(source="price.price_id")
    price_nickname = serializers.CharField(source="price.nickname")
    price = serializers.CharField(source="price.price")
    freq = serializers.CharField(source="price.freq")
    services = serializers.SerializerMethodField(method_name='get_feature_ids')
    subscription_status = serializers.CharField(source='subscription.status')
    period_start = serializers.DateTimeField(source='subscription.period_start')
    period_end = serializers.DateTimeField(source='subscription.period_end')
    trial_start = serializers.DateTimeField(source='subscription.trial_start')
    trial_end = serializers.DateTimeField(source='subscription.trial_end')
    ended_at = serializers.DateTimeField(source='subscription.ended_at')
    cancel_at = serializers.DateTimeField(source='subscription.cancel_at')
    cancel_at_period_end = serializers.BooleanField(source='subscription.cancel_at_period_end')

    def get_feature_ids(self, obj):
        return [{"feature_id": link.feature.feature_id, "feature_desc": link.feature.description} for link in
                obj.price.product.linked_features.all().prefetch_related('feature')]

    def get_subscription_expires_at(self, obj):
        return obj.subscription.period_end or \
               obj.subscription.cancel_at or \
               obj.subscription.trial_end or \
               obj.subscription.ended_at

    class Meta:
        model = SubscriptionItem
        fields = (
            "product_id", "product_name", "product_description", "price_id", "price_nickname", "price", "freq",
            "subscription_status", "period_start", "period_end",
            "trial_start", "trial_end", "ended_at", "cancel_at", "cancel_at_period_end", "services")


class ProductSerializer(serializers.ModelSerializer):
    class Meta:
        model = Product
        fields = "__all__"


class PriceSerializer(serializers.ModelSerializer):
    product_id = serializers.CharField(source="product.product_id")
    name = serializers.CharField(source="product.name")
    avail = serializers.BooleanField(source="active")
    services = serializers.SerializerMethodField(method_name='get_feature_ids')

    def get_feature_ids(self, obj):
        return [{"feature_id": prod_feature.feature.feature_id, "feature_desc": prod_feature.feature.description} for
                prod_feature in
                obj.product.linked_features.all().prefetch_related("feature")]

    class Meta:
        model = Price
        fields = ("price_id", "product_id", "name", "price", "freq", "avail", "services", "currency")


class CheckoutRequestSerializer(serializers.Serializer):
    """Handles request data to create a Stripe checkout session."""
    price_id = serializers.CharField()
    owner = serializers.HiddenField(default=serializers.CurrentUserDefault())

    def validate(self, attrs):
        stripe_user = get_or_create_stripe_user(user_id=self.context['request'].user.id)
        try:
            checkout_session = stripe_api_create_checkout_session(
                customer_id=stripe_user.customer_id,
                price_id=attrs['price_id'])
            attrs['session_id'] = checkout_session['id']
        except StripeError as e:
            raise ValidationError(e.error)
        return attrs

    def update(self, instance, validated_data):
        pass

    def create(self, validated_data):
        pass



# Generated by Django 4.0.1 on 2022-02-08 08:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('drf_stripe', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='description',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.AlterField(
            model_name='price',
            name='freq',
            field=models.CharField(blank=True, max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='price',
            name='nickname',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.AlterField(
            model_name='product',
            name='description',
            field=models.CharField(blank=True, max_length=1024, null=True),
        ),
        migrations.AlterField(
            model_name='product',
            name='name',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='cancel_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='ended_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='period_end',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='period_start',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='trial_end',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='trial_start',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from datetime import datetime
from typing import Optional, Dict, Union, List

from pydantic import BaseModel

from .currency import StripeCurrency
from .subscription import StripeSubscriptionItems


class StripeCustomer(BaseModel):
    """Based on https://stripe.com/docs/api/customers/object"""
    id: str
    address: Optional[Dict] = None
    description: Optional[str] = None
    email: Optional[str]
    metadata: Optional[Dict]
    name: Optional[str] = None
    phone: Optional[str] = None
    shipping: Optional[Dict] = None
    balance: Optional[int] = None
    created: Optional[datetime]
    currency: Optional[StripeCurrency] = None
    default_source: Optional[Union[str, Dict]] = None
    delinquent: Optional[bool]
    discount: Optional[Dict] = None
    invoice_prefix: Optional[str]
    invoice_settings: Optional[Dict]
    livemode: Optional[bool]
    next_invoice_sequence: Optional[int]
    preferred_locales: Optional[List[str]]
    sources: Optional[List[Dict]]
    subscriptions: Optional[List[StripeSubscriptionItems]]
    tax: Optional[Dict]
    tax_exempt: Optional[str]
    tax_ids: Optional[List[Dict]]


class StripeCustomers(BaseModel):
    """Based on https://stripe.com/docs/api/customers/list"""
    data: List[StripeCustomer]
    has_more: bool = None
    url: str = None



# Generated by Django 4.0.1 on 2022-01-13 15:48

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Feature',
            fields=[
                ('feature_id', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('description', models.CharField(max_length=256, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Price',
            fields=[
                ('price_id', models.CharField(max_length=256, primary_key=True, serialize=False)),
                ('nickname', models.CharField(max_length=256, null=True)),
                ('price', models.PositiveIntegerField()),
                ('freq', models.CharField(max_length=64, null=True)),
                ('active', models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('product_id', models.CharField(max_length=256, primary_key=True, serialize=False)),
                ('active', models.BooleanField()),
                ('description', models.CharField(max_length=1024, null=True)),
                ('name', models.CharField(max_length=256, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ProductFeature',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='StripeUser',
            fields=[
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True,
                                              related_name='stripe_user', serialize=False,
                                              to=settings.AUTH_USER_MODEL)),
                ('customer_id', models.CharField(max_length=128, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('subscription_id', models.CharField(max_length=256, primary_key=True, serialize=False)),
                ('period_start', models.DateTimeField(null=True)),
                ('period_end', models.DateTimeField(null=True)),
                ('cancel_at', models.DateTimeField(null=True)),
                ('cancel_at_period_end', models.BooleanField()),
                ('ended_at', models.DateTimeField(null=True)),
                ('status', models.CharField(max_length=64)),
                ('trial_end', models.DateTimeField(null=True)),
                ('trial_start', models.DateTimeField(null=True)),
                ('stripe_user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions',
                                   to='drf_stripe.stripeuser')),
            ],
        ),
        migrations.CreateModel(
            name='SubscriptionItem',
            fields=[
                ('sub_item_id', models.CharField(max_length=256, primary_key=True, serialize=False)),
                ('quantity', models.PositiveIntegerField()),
                ('price', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+',
                                            to='drf_stripe.price')),
                ('subscription', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items',
                                                   to='drf_stripe.subscription')),
            ],
        ),
        migrations.AddIndex(
            model_name='stripeuser',
            index=models.Index(fields=['user', 'customer_id'], name='drf_stripe__user_id_6bbc0d_idx'),
        ),
        migrations.AddField(
            model_name='productfeature',
            name='feature',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='linked_products',
                                    to='drf_stripe.feature'),
        ),
        migrations.AddField(
            model_name='productfeature',
            name='product',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='linked_features',
                                    to='drf_stripe.product'),
        ),
        migrations.AddField(
            model_name='price',
            name='product',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='prices',
                                    to='drf_stripe.product'),
        ),
        migrations.AddIndex(
            model_name='subscription',
            index=models.Index(fields=['stripe_user', 'status'], name='drf_stripe__stripe__7d71e7_idx'),
        ),
        migrations.AddIndex(
            model_name='price',
            index=models.Index(fields=['active', 'freq'], name='drf_stripe__active_3854b4_idx'),
        ),
    ]



# Generated by Django 4.0.2 on 2022-02-08 18:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('drf_stripe', '0002_alter_feature_description_alter_price_freq_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='price',
            name='currency',
            field=models.CharField(default='usd', max_length=3),
            preserve_default=False,
        ),
    ]
